@@ -1,6 +1,6 @@
 # utils for working with 3d-protein structures
 import os
-import numpy as np 
+import numpy as np
 import torch
 from functools import wraps
 from einops import rearrange, repeat
@@ -8,13 +8,17 @@ from einops import rearrange, repeat
 # bio
 import mdtraj
 try:
-    from sidechainnet.sequence.utils import VOCAB
+    from sidechainnet.utils.sequence import ProteinVocabulary as VOCAB
     from sidechainnet.utils.measure import GLOBAL_PAD_CHAR
-    from sidechainnet.structure.build_info import NUM_COORDS_PER_RES, SC_BUILD_INFO
+    from sidechainnet.structure.build_info import NUM_COORDS_PER_RES, BB_BUILD_INFO, SC_BUILD_INFO
     from sidechainnet.structure.StructureBuilder import _get_residue_build_iter
+    # build vocablary
+    VOCAB = VOCAB()
 except:
     NUM_COORDS_PER_RES = 14
     GLOBAL_PAD_CHAR = 0
+    BB_BUILD_INFO = {"BONDLENS": {"c-o": 1.229},
+                     "BONDANGS": {"ca-c-o": 2.0944}}
 
 # own
 
@@ -35,10 +39,10 @@ def set_backend_kwarg(fn):
         return fn(*args, **kwargs)
     return inner
 
-def expand_dims_to(t, len = 3):
-    if len == 0:
+def expand_dims_to(t, length = 3):
+    if length == 0:
         return t
-    return t.reshape(*((1,) * len), *t.shape) # will work with both torch and numpy
+    return t.reshape(*((1,) * length), *t.shape) # will work with both torch and numpy
 
 def expand_arg_dims(dim_len = 3):
     """ pack here for reuse. 
@@ -48,9 +52,9 @@ def expand_arg_dims(dim_len = 3):
         @wraps(fn)
         def inner(x, y, **kwargs):
             assert len(x.shape) == len(y.shape), "Shapes of A and B must match."
-            remaining_len = len(x.shape) - dim_len
-            x = expand_dims_to(x, remaining_len)
-            y = expand_dims_to(y, remaining_len)
+            remaining_len = dim_len - len(x.shape)
+            x = expand_dims_to(x, length = remaining_len)
+            y = expand_dims_to(y, length = remaining_len)
             return fn(x, y, **kwargs)
         return inner
     return outer
@@ -60,7 +64,12 @@ def invoke_torch_or_numpy(torch_fn, numpy_fn):
         @wraps(fn)
         def inner(*args, **kwargs):
             backend = kwargs.pop('backend')
-            passed_args, passed_kwargs = fn(*args, **kwargs)
+            passed_args = fn(*args, **kwargs)
+            passed_args = list(passed_args)
+            if isinstance(passed_args[-1], dict):
+                passed_kwargs = passed_args.pop()
+            else:
+                passed_kwargs = {}
             backend_fn = torch_fn if backend == 'torch' else numpy_fn
             return backend_fn(*passed_args, **passed_kwargs)
         return inner
@@ -142,74 +151,108 @@ def custom2pdb(coords, proteinnet_id, route):
 
 # sidechainnet utils
 
-def scn_cloud_mask(scn_seq, bool=True):
+def scn_cloud_mask(scn_seq, boolean=True):
     """ Gets the boolean mask atom positions (not all aas have same atoms). 
         Inputs: 
         * scn_seq: (batch, length) sequence as provided by Sidechainnet package
-        * bool: whether to return as array of idxs or boolean values
-        * mask: (batch, length). current mask to build a custom chain_mask 
+        * boolean: whether to return as array of idxs or boolean values
         Outputs: (batch, length, NUM_COORDS_PER_RES) boolean mask 
     """
     # scaffolds 
-    mask = torch.zeros(*scn_seq.shape, NUM_COORDS_PER_RES, device=snc_seq.device)
-    chain_mask = []
+    mask = torch.zeros(*scn_seq.shape, NUM_COORDS_PER_RES, device=scn_seq.device)
     # fill 
-    for n in range(len(masks)):
-        for i,aa in enumerate(scn_seq.cpu().numpy()):
+    for n, seq in enumerate(scn_seq.cpu().numpy()):
+        for i,aa in enumerate(seq):
             # get num of atom positions - backbone is 4: ...N-C-C(=O)...
-            n_atoms = 4+len( SC_BUILD_INFO[VOCAB.int2chars(x)]["atom-names"] )
+            n_atoms = 4+len( SC_BUILD_INFO[VOCAB.int2chars(aa)]["atom-names"] )
             mask[n, i, :n_atoms] = 1
-    return mask.bool()
+    if boolean:
+        return mask.bool()
+    return mask.nonzero()
 
-def scn_backbone_mask(scn_seq, bool=True, l_aa=NUM_COORDS_PER_RES):
+def scn_backbone_mask(scn_seq, boolean=True, l_aa=NUM_COORDS_PER_RES):
     """ Gets the boolean mask for N and CA positions. 
         Inputs: 
         * scn_seq: sequence as provided by Sidechainnet package
         * bool: whether to return as array of idxs or boolean values
         Outputs: (N_mask, CA_mask)
     """
-    lengths = np.arange(scn_seq.shape[-1]*l_aa)
+    lengths = torch.arange(scn_seq.shape[-1]*l_aa)
+    # repeat if needed:
+    if len(lengths.shape) == 2:
+        lengths = repeat(lengths, 'l -> b l', b=scn_seq.shape[0])
     # N is the first atom in every AA. CA is the 2nd.
     N_mask  = lengths%l_aa == 0
     CA_mask = lengths%l_aa == 1
     if boolean:
         return N_mask, CA_mask
+    return N_mask.nonzero(), CA_mask.nonzero()
 
-    return lengths[N_mask], lengths[CA_mask]
+def nerf_torch(a, b, c, l, theta, chi):
+    """ Custom Natural extension of Reference Frame. 
+        Inputs:
+        * a: (batch, 3) or (3,). point(s) of the plane, not connected to d
+        * b: (batch, 3) or (3,). point(s) of the plane, not connected to d
+        * c: (batch, 3) or (3,). point(s) of the plane, connected to d
+        * theta: (batch,) or (float).  angle(s) between b-c-d
+        * chi: (batch,) or float. dihedral angle(s) between the a-b-c and b-c-d planes
+        Outputs: d (batch, 3) or (3,). the next point in the sequence, linked to c
+    """
+    # safety check
+    if not ( (-np.pi <= theta) * (theta <= np.pi) ).all().item():
+        raise ValueError(f"theta(s) must be in radians and in [-pi, pi]. theta(s) = {theta}")
+    # calc vecs
+    ba = b-a
+    cb = c-b
+    # calc rotation matrix. based on plane normals and normalized
+    n_plane  = torch.cross(ba, cb, dim=-1)
+    n_plane_ = torch.cross(n_plane, cb, dim=-1)
+    rotate   = torch.stack([cb, n_plane_, n_plane], dim=-1)
+    rotate  /= torch.norm(rotate, dim=-2, keepdim=True)
+    # calc proto point, rotate
+    d = torch.stack([torch.cos(theta),
+                     torch.sin(theta) * torch.cos(chi),
+                     torch.sin(theta) * torch.sin(chi)], dim=-1).unsqueeze(-1)
+    # extend base point, set length
+    return c + l.unsqueeze(-1) * torch.matmul(rotate, d).squeeze()
 
-def sidechain_3d(seq, backbone, n_atoms=NUM_COORDS_PER_RES, 
-                 padding=GLOBAL_PAD_CHAR, force=False):
+def sidechain_container(backbones, place_oxygen=False,
+                        n_atoms=NUM_COORDS_PER_RES, padding=GLOBAL_PAD_CHAR):
     """ Gets a backbone of the protein, returns the whole coordinates
         with sidechains (same format as sidechainnet). Keeps differentiability.
         Inputs: 
-        * seq: (L,) tensor of ints. sequence tokens.
-        * backbone: (batch, L*3, 3): assume batch=1 (could be extended later).
+        * backbones: (batch, L*3, 3): assume batch=1 (could be extended later).
                     Coords for (N-term, C-alpha, C-term) of every aa.
+        * place_oxygen: whether to claculate the oxygen of the
+                        carbonyl group via NeRF
         * n_atoms: int. n of atom positions / atom. same as in sidechainnet: 14
         * padding: int. padding token. same as in sidechainnet: 0
         Outputs: whole coordinates of shape (batch, L, n_atoms, 3)
     """
-    batch, length = list(backbone.shape[:2])
-    new_coords = torch.ones(batch, length//3, NUM_COORDS_PER_RES, 3) * padding
-    new_coords[:, :, :3] = rearrange(backbone, 'b (l back) d -> b l back d', back=3)
-    # build sidechain for every aa
-    for i,token in enumerate(seq):
-        # position C-beta
-        # 
-        # # get location of (=O) spanding from previous C-term
-        #
-        # # get tetrahedral conformation of C-alpha, find closest to (=O), exploit D-aa
-        #
-        # iterate over aa atoms and place them accordingly
-        for j, (bond_len, angle, torsion, atom_names) in enumerate(
-            _get_residue_build_iter(token, SC_BUILD_INFO)):
-            pass 
-            # extend from previous atom
-            #
-    if force:
-        return new_coords
+    batch, length = backbones.shape[0], backbones.shape[1] // 3
+    new_coords = torch.ones(batch, length, NUM_COORDS_PER_RES, 3, device=backbones.device) * padding
+    new_coords[:, :, :3] = rearrange(backbones, 'b (l back) d -> b l back d', l=length)
+    # set the rest of positions to c_alpha
+    new_coords[:, :, 3:] = repeat(new_coords[:, :, 2], 'b l d -> b l scn d', scn=11)
+    # hard-calculate oxygen position of carbonyl group
+    if place_oxygen: 
+        # build (=O) position of revery aa in each chain
+        for s in range(batch):
+            # dihedrals phi=f(c-1, n, ca, c) & psi=f(n, ca, c, n+1)
+            # phi = get_dihedral_torch(*backbone[s, i*3 - 1 : i*3 + 3]) if i>0 else None
+            psis = torch.tensor([ get_dihedral_torch(*backbones[s, i*3 + 0 : i*3 + 4] )if i < length-1 else np.pi*5/4 \
+                                  for i in range(length) ])
+            # the angle for placing oxygen is opposite to psi of current res.
+            # psi not available for last one so pi/4 taken for now
+            bond_lens  = repeat(torch.tensor(BB_BUILD_INFO["BONDLENS"]["c-o"]), ' -> b', b=length).to(psis.device)
+            bond_angs  = repeat(torch.tensor(BB_BUILD_INFO["BONDANGS"]["ca-c-o"]), ' -> b', b=length).to(psis.device)
+            correction = repeat(torch.tensor(-np.pi), ' -> b', b=length).to(psis.device) 
+            new_coords[:, :, 3] = nerf_torch(new_coords[:, :, 0], 
+                                             new_coords[:, :, 1], 
+                                             new_coords[:, :, 2], 
+                                             bond_lens, bond_angs, psis + correction)
+    return new_coords
 
-    raise NotImplementedError("Function not implemented yet")
 
 
 # distance utils (distogram to dist mat + masking)
@@ -227,34 +270,35 @@ def center_distogram_torch(distogram, bins=DISTANCE_THRESHOLDS, min_t=1., center
     """
     shape, device = distogram.shape, distogram.device
     # threshold to weights and find mean value of each bin
-    n_bins = bins - 0.5 * (bins[2] - bins[1])
+    n_bins = ( bins - 0.5 * (bins[2] - bins[1]) ).to(device)
     n_bins[0]  = 1.5
-    # TODO: adapt so that mean option considers IGNORE_INDEX
-    n_bins[-1] = n_bins[-1]
-    n_bins = n_bins.to(device)
+    n_bins[-1] = 1.33*bins[-1] # above last threshold is ignored
+    max_bin_allowed = torch.tensor(n_bins.shape[0]-1).to(device).long()
     # calculate measures of centrality and dispersion - 
     if center == "median":
         cum_dist = torch.cumsum(distogram, dim=-1)
-        medium   = 0.5 * torch.ones(*cum_dist.shape[:-1], device=device).unsqueeze(dim=-1)
+        medium   = 0.5 * torch.ones(*cum_dist.shape[:-1], 1, device=device)
         central  = torch.searchsorted(cum_dist, medium).squeeze()
-        central  = n_bins[ torch.minimum(central, torch.tensor( len(n_bins)-1 )).long() ].unsqueeze(dim=0)
+        central  = n_bins[ torch.min(central, max_bin_allowed) ]
     elif center == "mean":
         central  = (distogram * n_bins).sum(dim=-1)
     # create mask for last class - (IGNORE_INDEX)   
     mask = (central <= bins[-2].item()).float()
     # mask diagonal to 0 dist 
     diag = np.arange(shape[-2])
+    central = expand_dims_to(central, 3 - len(central.shape))
     central[:, diag, diag] = 0.
     # provide weights
     if wide == "var":
         dispersion = (distogram * (n_bins - central.unsqueeze(-1))**2).sum(dim=-1)
     elif wide == "std":
+
         dispersion = (distogram * (n_bins - central.unsqueeze(-1))**2).sum(dim=-1).sqrt()
     else:
         dispersion = torch.zeros_like(central, device=central.device)
-    # rescale to 0-1. lower std / var  --> weight=1
+    # rescale to 0-1. lower std / var  --> weight=1. set potential nan's to 0
     weights = mask / (1+dispersion)
-
+    weights[weights != weights] = 0.
     return central, weights
 
 # distance matrix to 3d coords: https://github.com/scikit-learn/scikit-learn/blob/42aff4e2e/sklearn/manifold/_mds.py#L279
@@ -271,8 +315,8 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
     if weights is None:
         weights = torch.ones_like(pre_dist_mat)
 
-    # batched MDS
-    pre_dist_mat = expand_dims_to(pre_dist_mat, len = (len(pre_dist_mat) - 3))
+    # ensure batched MDS
+    pre_dist_mat = expand_dims_to(pre_dist_mat, length = ( 3 - len(pre_dist_mat.shape) ))
 
     # start
     batch, N, _ = pre_dist_mat.shape
@@ -284,11 +328,11 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
     for i in range(iters):
         # compute distance matrix of coords and stress
         dist_mat = torch.cdist(best_3d_coords, best_3d_coords, p=2)
-        stress   = ( weights * (dist_mat - pre_dist_mat)**2 ).sum(dim=(-1,-2)) / 2
+        stress   = ( weights * (dist_mat - pre_dist_mat)**2 ).sum(dim=(-1,-2)) * 0.5
         # perturb - update X using the Guttman transform - sklearn-like
         dist_mat[dist_mat == 0] = 1e-7
         ratio = weights * (pre_dist_mat / dist_mat)
-        B = ratio * (-1)
+        B = -ratio
         B[:, np.arange(N), np.arange(N)] += ratio.sum(dim=-1)
         # update
         coords = (1. / N * torch.matmul(B, best_3d_coords))
@@ -299,7 +343,7 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
         if (best_stress - stress / dis).mean() <= tol:
             if verbose:
                 print('breaking at iteration %d with stress %s' % (i,
-                                                                   stress))
+                                                                   stress / dis))
             break
 
         best_3d_coords = coords
@@ -326,11 +370,11 @@ def mds_numpy(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
     for i in range(iters):
         # compute distance matrix of coords and stress
         dist_mat = np.linalg.norm(np.expand_dims(best_3d_coords,1) - np.expand_dims(best_3d_coords,2), axis=0)
-        stress   = (( weights * (dist_mat - pre_dist_mat) )**2).sum() / 2
+        stress   = (( weights * (dist_mat - pre_dist_mat) )**2).sum() * 0.5
         # perturb - update X using the Guttman transform - sklearn-like
         dist_mat[dist_mat == 0] = 1e-7
         ratio = pre_dist_mat / dist_mat
-        B = ratio * (-1)
+        B = -ratio 
         B[np.arange(N), np.arange(N)] += ratio.sum(axis=1)
         # update - double transpose. TODO: consider fix
         coords = (1. / N * np.dot(best_3d_coords, B))
@@ -341,7 +385,7 @@ def mds_numpy(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
         if (best_stress - stress / dis) <= tol:
             if verbose:
                 print('breaking at iteration %d with stress %s' % (i,
-                                                                   stress))
+                                                                   stress / dis))
             break
 
         best_3d_coords = coords
@@ -350,50 +394,66 @@ def mds_numpy(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
 
     return best_3d_coords, np.array(his)
 
-def get_dihedral_torch(c1, c2, c3, c4, c5):
+def get_dihedral_torch(c1, c2, c3, c4):
     """ Returns the dihedral angle in radians.
         Will use atan2 formula from: 
         https://en.wikipedia.org/wiki/Dihedral_angle#In_polymer_physics
+        Can't use torch.dot bc it does not broadcast
+        Inputs: 
+        * c1: (batch, 3) or (3,)
+        * c1: (batch, 3) or (3,)
+        * c1: (batch, 3) or (3,)
+        * c1: (batch, 3) or (3,)
     """
-
     u1 = c2 - c1
     u2 = c3 - c2
     u3 = c4 - c3
-    u4 = c5 - c4
 
-    return torch.atan2( torch.dot( torch.norm(u2) * u1, torch.cross(u3,u4) ),  
-                        torch.dot( torch.cross(u1,u2), torch.cross(u3, u4) ) ) 
+    return torch.atan2( ( (torch.norm(u2, dim=-1, keepdim=True) * u1) * torch.cross(u2,u3, dim=-1) ).sum(dim=-1) ,  
+                        (  torch.cross(u1,u2, dim=-1) * torch.cross(u2, u3, dim=-1) ).sum(dim=-1) ) 
 
-def get_dihedral_numpy(c1, c2, c3, c4, c5):
+
+def get_dihedral_numpy(c1, c2, c3, c4):
     """ Returns the dihedral angle in radians.
         Will use atan2 formula from: 
         https://en.wikipedia.org/wiki/Dihedral_angle#In_polymer_physics
+        Inputs: 
+        * c1: (batch, 3) or (3,)
+        * c1: (batch, 3) or (3,)
+        * c1: (batch, 3) or (3,)
+        * c1: (batch, 3) or (3,)
     """
-
     u1 = c2 - c1
     u2 = c3 - c2
     u3 = c4 - c3
-    u4 = c5 - c4
 
-    return np.arctan2( np.dot( np.linalg.norm(u2) * u1, np.cross(u3,u4) ),  
-                       np.dot( np.cross(u1,u2), np.cross(u3, u4) ) ) 
+    return np.arctan2( ( (np.linalg.norm(u2, axis=-1, keepdims=True) * u1) * np.cross(u2,u3, axis=-1)).sum(axis=-1),  
+                       ( np.cross(u1,u2, axis=-1) * np.cross(u2, u3, axis=-1) ).sum(axis=-1) ) 
 
-def fix_mirrors_torch(preds, stresses, N_mask, CA_mask, verbose=0):
+def fix_mirrors_torch(preds, stresses, N_mask, CA_mask, C_mask=None, verbose=0):
     """ Filters mirrors selecting the 1 with most N of negative phis.
         Used as part of the MDScaling wrapper if arg is passed. See below.
-        Angle Phi between planes: (Ca{-1}, N, Ca{0}) and (Ca{0}, N{+1}, C_a{+1})
+        Angle Phi between planes: (Cterm{-1}, N, Ca{0}) and (N{0}, Ca{+1}, Cterm{+1})
+        * preds: (n_mirrors, 3, N)
+        * stresses: (n_mirrors, steps) historic stresses
+        * N_mask: (N, ) boolean mask for N-term positions
+        * CA_mask: (N, ) boolean mask for C-alpha positions
+        * C_mask: (N, ) or None. boolean mask for C-alpha positions or
+                    automatically calculate from N_mask and CA_mask if None.
+        * verbose: bool. verbosity level
     """ 
-    preds_ = preds.detach()
-    ns = torch.transpose(preds_, -1, -2)[:, N_mask][:, 1:]
-    cs = torch.transpose(preds_, -1, -2)[:, CA_mask]
+    # detach gradients for angle calculation - mirror selection
+    preds_ = torch.transpose(preds.detach(), -1 , -2).cpu()
+    n_terms  = preds_[:, N_mask.squeeze()]
+    c_alphas = preds_[:, CA_mask.squeeze()]
+    # select c_term auto if not passed
+    if C_mask is not None: 
+        c_terms = preds_[:, C_mask]
+    else:
+        c_terms  = preds_[ :, torch.logical_not(torch.logical_or(N_mask,CA_mask)).squeeze() ]
     # compute phis and count lower than 0s
-    phis_count = []
-    for i in range(cs.shape[0]):
-        # calculate phi angles
-        phis = [ get_dihedral_torch(cs[i,j-1], ns[i,j], cs[i,j], ns[i,j+1], cs[i,j+1]) \
-                 for j in range(1, cs.shape[1]-1) ]
-
-        phis_count.append( (np.array(phis)<0).sum() )
+    phis_count = [ (get_dihedral_torch(c_terms[i,:-1], n_terms[i, 1:], c_alphas[i, 1:], c_terms[i, 1:])<0).sum().item() \
+                   for i in range(preds.shape[0])]
 
     idx = np.argmax(phis_count)
     # debugging/testing if arg passed
@@ -401,26 +461,34 @@ def fix_mirrors_torch(preds, stresses, N_mask, CA_mask, verbose=0):
         print("Negative phis:", phis_count, "selected", idx)
     return preds[idx].unsqueeze(0), stresses[idx]
 
-def fix_mirrors_numpy(preds, stresses, N_mask, CA_mask, verbose=0):
+def fix_mirrors_numpy(preds, stresses, N_mask, CA_mask, C_mask=None, verbose=0):
     """ Filters mirrors selecting the 1 with most N of negative phis.
         Used as part of the MDScaling wrapper if arg is passed. See below.
-        Angle Phi between planes: (Ca{-1}, N, Ca{0}) and (Ca{0}, N{+1}, C_a{+1})
+        Angle Phi between planes: (Cterm{-1}, N, Ca{0}) and (N{0}, Ca{+1}, Cterm{+1})
+        * preds: (n_mirrors, 3, N)
+        * stresses: (n_mirrors, steps) historic stresses
+        * N_mask: (N, ) boolean mask for N-term positions
+        * CA_mask: (N, ) boolean mask for C-alpha positions
+        * C_mask: (N, ) or None. boolean mask for C-alpha positions or
+                    automatically calculate from N_mask and CA_mask if None.
+        * verbose: bool. verbosity level
     """ 
-    ns = np.transpose(preds, (0, 2, 1))[:, N_mask][:, 1:]
-    cs =  np.transpose(preds, (0, 2, 1))[:, CA_mask]
-    # compute phis and count lower than 0s
-    phis_count = []
-    for i in range(cs.shape[0]):
-        # calculate phi angles
-        phis = [ get_dihedral_numpy(cs[i,j-1], ns[i,j], cs[i,j], ns[i,j+1], cs[i,j+1]) \
-                 for j in range(1, cs.shape[1]-1) ]
-
-        phis_count.append( (np.array(phis)<0).sum() )
+    preds_ = np.transpose(preds, (0, 2, 1))
+    n_terms  = preds_[:, N_mask.squeeze()] 
+    c_alphas = preds_[:, CA_mask.squeeze()]
+    # select c_term auto if not passed
+    if C_mask is not None: 
+        c_terms = preds_[:, C_mask]
+    else:
+        c_terms  = preds_[:, (np.ones_like(N_mask)-N_mask-CA_mask).squeeze().astype(bool) ]
+    # compute number of phis lower than 0
+    phis_count = [ (get_dihedral_numpy(c_terms[i,:-1], n_terms[i,:], c_alphas[i,:], c_terms[i,:])<0).sum() \
+                   for i in range(preds.shape[0])]
 
     idx = np.argmax(phis_count)
     # debugging/testing if arg passed
     if verbose:
-        print("Negative phis:", phis_count)
+        print("Negative phis:", phis_count, "selected", idx)
     return preds[idx], stresses[idx]
 
 
@@ -489,12 +557,13 @@ def gdt_torch(X, Y, cutoffs, weights=None):
         * cutoffs is a list of `K` thresholds
         * weights is a list of `K` weights (1 x each threshold)
     """
+    device = X.device
     if weights is None:
         weights = torch.ones(1,len(cutoffs))
     else:
-        weights = torch.tensor([weights]).to(x.device)
+        weights = torch.tensor([weights]).to(device)
     # set zeros and fill with values
-    GDT = torch.zeros(X.shape[0], len(cutoffs), device=X.device) 
+    GDT = torch.zeros(X.shape[0], len(cutoffs), device=device)
     dist = ((X - Y)**2).sum(dim=1).sqrt()
     # iterate over thresholds
     for i,cutoff in enumerate(cutoffs):
@@ -540,19 +609,19 @@ def tmscore_numpy(X, Y):
 
 
 def mdscaling_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5,
-              fix_mirror=0, N_mask=None, CA_mask=None, verbose=2):
+              fix_mirror=0, N_mask=None, CA_mask=None, C_mask=None, verbose=2):
     # repeat for mirrors calculations
-    pre_dist_mat = repeat(pre_dist_mat, 'ni nj -> m ni nj', m = max(1,fix_mirror))
+    pre_dist_mat = repeat(pre_dist_mat, '() ni nj -> m ni nj', m = max(1, fix_mirror))
     # batched mds for full parallel 
     preds, stresses = mds_torch(pre_dist_mat, weights=weights,iters=iters, 
                                               tol=tol, verbose=verbose)
     if not fix_mirror:
         return preds[0], stresses[0]
 
-    return fix_mirrors_torch(preds, stresses, N_mask, CA_mask)
+    return fix_mirrors_torch(preds, stresses, N_mask, CA_mask, C_mask)
 
 def mdscaling_numpy(pre_dist_mat, weights=None, iters=10, tol=1e-5,
-              fix_mirror=0, N_mask=None, CA_mask=None, verbose=2):
+              fix_mirror=0, N_mask=None, CA_mask=None, C_mask=None, verbose=2):
     preds = [mds_numpy(pre_dist_mat, weights=weights,iters=iters, 
                                          tol=tol, verbose=verbose) \
                  for i in range( max(1,fix_mirror) )]
@@ -561,7 +630,7 @@ def mdscaling_numpy(pre_dist_mat, weights=None, iters=10, tol=1e-5,
         return preds[0]
 
     return fix_mirrors_numpy([x[0] for x in preds],
-                             [x[1] for x in preds], N_mask, CA_mask)
+                             [x[1] for x in preds], N_mask, CA_mask, C_mask)
 ################
 ### WRAPPERS ###
 ################
@@ -638,7 +707,7 @@ def GDT(A, B, *, mode="TS", cutoffs=[1,2,4,8], weights=None):
     # define cutoffs for each type of gdt and weights
     cutoffs = [0.5,1,2,4] if mode in ["HA", "ha"] else [1,2,4,8]
     # calculate GDT
-    return (A, B, cutoffs), {'weights': weights}
+    return A, B, cutoffs, {'weights': weights}
 
 @expand_arg_dims()
 @set_backend_kwarg
